use std::hash::Hash;

use anyhow::{anyhow, bail};
use i2p::net::{I2pAddr, I2pSocketAddr, ToI2pSocketAddrs};
use serde::{Deserialize, Serialize};

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, PartialOrd, Ord)]
pub struct PeerId {
    i2p_dest: I2pSocketAddr,
    i2p_b32: Option<I2pAddr>,
}

impl PeerId {
    pub fn addr(&self) -> I2pSocketAddr {
        self.i2p_dest.to_owned()
    }
    pub fn addr_ref(&self) -> &I2pSocketAddr {
        &self.i2p_dest
    }
    pub fn b32_addr(&mut self) -> anyhow::Result<I2pAddr> {
        let result = I2pAddr::from_b64(&self.i2p_dest.dest().string());
        if let Ok(addr) = &result {
            self.i2p_b32 = Some(addr.to_owned());
        }
        result.map_err(|e| anyhow!(e))
    }
    pub fn b32_addr_nocache(&self) -> anyhow::Result<I2pAddr> {
        I2pAddr::from_b64(&self.i2p_dest.dest().string()).map_err(|e| anyhow!(e))
    }
}

// The identity of the PeerId only depends on the i2p_dest (which is unique),
// and not on whether the b32 address has been computed before
impl Hash for PeerId {
    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
        self.i2p_dest.hash(state);
    }
}

impl ToString for PeerId {
    fn to_string(&self) -> String {
        self.i2p_dest.to_string()
    }
}

impl TryFrom<&str> for PeerId {
    type Error = anyhow::Error;

    fn try_from(value: &str) -> Result<Self, anyhow::Error> {
        match ToI2pSocketAddrs::to_socket_addrs(&value) {
            Ok(addr_iter) => {
                for addr in addr_iter {
                    return Ok(PeerId { i2p_dest: addr, i2p_b32: None });
                }
                return Err(anyhow::Error::msg("No valid I2P address found"));
            }
            Err(e) => bail!(e),
        }
    }
}

impl TryFrom<String> for PeerId {
    type Error = anyhow::Error;

    fn try_from(value: String) -> Result<Self, Self::Error> {
        Self::try_from(value.as_str())
    }
}

impl From<I2pSocketAddr> for PeerId {
    fn from(value: I2pSocketAddr) -> Self {
        PeerId { i2p_dest: value, i2p_b32: None }
    }
}

impl From<PeerId> for I2pSocketAddr {
    fn from(value: PeerId) -> Self {
        value.i2p_dest
    }
}

impl Default for PeerId {
    fn default() -> Self {
        PeerId {
            i2p_dest: I2pSocketAddr::new(I2pAddr::new(""), 0),
            i2p_b32: None
        }
    }
}
